import os
import shutil
import tempfile
import unittest
from pathlib import Path
from tempfile import TemporaryDirectory

import petl

from parsons import Table
from parsons.utilities import zip_archive
from test.utils import assert_matching_tables

# Notes :
# - The `Table.to_postgres()` test is housed in the Postgres tests
# - The `Table.from_postgres()` test is housed in the Postgres test


class TestParsonsTable(unittest.TestCase):
    def setUp(self):
        # Create Table object
        self.lst = [
            {"a": 1, "b": 2, "c": 3},
            {"a": 4, "b": 5, "c": 6},
            {"a": 7, "b": 8, "c": 9},
            {"a": 10, "b": 11, "c": 12},
            {"a": 13, "b": 14, "c": 15},
        ]
        self.lst_dicts = [{"first": "Bob", "last": "Smith"}]
        self.tbl = Table(self.lst_dicts)

        # Create a tmp dir
        self.tmp_folder = tempfile.mkdtemp()

    def tearDown(self):
        # Delete tmp folder and files
        shutil.rmtree(self.tmp_folder)

    def test_from_list_of_dicts(self):
        tbl = Table(self.lst)

        # Test Iterate and is list like
        self.assertEqual(tbl[0], {"a": 1, "b": 2, "c": 3})

    def test_from_list_of_lists(self):
        list_of_lists = [
            ["a", "b", "c"],
            [1, 2, 3],
            [4, 5, 6],
        ]
        tbl = Table(list_of_lists)

        self.assertEqual(tbl[0], {"a": 1, "b": 2, "c": 3})

    def test_from_petl(self):
        nrows = 10
        ptbl = petl.dummytable(numrows=nrows)
        tbl = Table(ptbl)
        self.assertEqual(tbl.num_rows, nrows)

    def test_from_invalid_list(self):
        # Tests that a table can't be created from a list of invalid items
        list_of_invalid = [1, 2, 3]
        self.assertRaises(ValueError, Table, list_of_invalid)

    def test_from_empty_petl(self):
        # This test ensures that this would fail: Table(None)
        # Even while allowing Table() to work
        self.assertRaises(ValueError, Table, None)

    def test_from_empty_list(self):
        # Just ensure this doesn't throw an error
        Table()
        Table([])
        Table([[]])

    def test_materialize(self):
        # Simple test that materializing doesn't change the table
        tbl_materialized = Table(self.lst_dicts)
        tbl_materialized.materialize()

        assert_matching_tables(self.tbl, tbl_materialized)

    def test_materialize_to_file(self):
        # Simple test that materializing doesn't change the table
        tbl_materialized = Table(self.lst_dicts)
        _ = tbl_materialized.materialize_to_file()

        assert_matching_tables(self.tbl, tbl_materialized)

    def test_empty_column(self):
        # Test that returns True on an empty column and False on a populated one.

        tbl = Table([["a", "b"], ["1", None], ["2", None]])

        self.assertTrue(tbl.empty_column("b"))
        self.assertFalse(tbl.empty_column("a"))

    def test_from_columns(self):
        header = ["col1", "col2"]
        col1 = [1, 2, 3]
        col2 = ["a", "b", "c"]
        tbl = Table.from_columns([col1, col2], header=header)

        self.assertEqual(tbl[0], {"col1": 1, "col2": "a"})

    # Removing this test since it is an optional dependency.
    """
    def test_from_datafame(self):

        import pandas

        # Assert creates table without index
        tbl = Table(self.lst)
        tbl_from_df = Table.from_dataframe(tbl.to_dataframe())
        assert_matching_tables(tbl, tbl_from_df)


    def test_to_dataframe(self):

        # Is a dataframe
        self.assertIsInstance(self.tbl.to_dataframe(), pandas.core.frame.DataFrame)
    """

    def test_to_petl(self):
        # Is a petl table
        self.assertIsInstance(self.tbl.to_petl(), petl.io.json.DictsView)

    def test_to_html(self):
        html_file = str(Path(self.tmp_folder) / "test.html")

        # Test writing file
        self.tbl.to_html(html_file)

        # Test written correctly
        html = (
            "<table class='petl'>\n"
            "<thead>\n"
            "<tr>\n"
            "<th>first</th>\n"
            "<th>last</th>\n"
            "</tr>\n"
            "</thead>\n"
            "<tbody>\n"
            "<tr>\n"
            "<td>Bob</td>\n"
            "<td>Smith</td>\n"
            "</tr>\n"
            "</tbody>\n"
            "</table>\n"
        )
        with open(html_file, "r") as f:
            self.assertEqual(f.read(), html)

    def test_to_temp_html(self):
        # Test write to object
        path = self.tbl.to_html()

        # Written correctly
        html = (
            "<table class='petl'>\n"
            "<thead>\n"
            "<tr>\n"
            "<th>first</th>\n"
            "<th>last</th>\n"
            "</tr>\n"
            "</thead>\n"
            "<tbody>\n"
            "<tr>\n"
            "<td>Bob</td>\n"
            "<td>Smith</td>\n"
            "</tr>\n"
            "</tbody>\n"
            "</table>\n"
        )
        with open(path, "r") as f:
            self.assertEqual(f.read(), html)

    def _assert_expected_csv(self, path, orig_tbl):
        result_tbl = Table.from_csv(path)
        assert_matching_tables(orig_tbl, result_tbl)

    def test_to_from_csv(self):
        with TemporaryDirectory() as tempdir:
            path = str(Path(tempdir) / "test.csv")
            self.tbl.to_csv(path)
            self._assert_expected_csv(path, self.tbl)

    def test_to_from_csv_compressed(self):
        with TemporaryDirectory() as tempdir:
            path = str(Path(tempdir) / "test.csv.gz")
            self.tbl.to_csv(path)
            self._assert_expected_csv(path, self.tbl)

    def test_to_from_temp_csv(self):
        path = self.tbl.to_csv()
        self._assert_expected_csv(path, self.tbl)

    def test_to_from_temp_csv_compressed(self):
        path = self.tbl.to_csv(temp_file_compression="gzip")
        self._assert_expected_csv(path, self.tbl)

    def test_from_csv_string(self):
        path = self.tbl.to_csv()
        # Pull the file into a string
        with open(path, "r") as f:
            csv_string = f.read()

        result_tbl = Table.from_csv_string(csv_string)
        assert_matching_tables(self.tbl, result_tbl)

    def test_append_csv_compressed(self):
        path = self.tbl.to_csv(temp_file_compression="gzip")
        append_tbl = Table([{"first": "Mary", "last": "Nichols"}])
        append_tbl.append_csv(path)

        result_tbl = Table.from_csv(path)
        # Combine tables, so we can check the resulting file
        self.tbl.concat(append_tbl)
        assert_matching_tables(self.tbl, result_tbl)

    def test_from_csv_raises_on_empty_file(self):
        # Create empty file
        with TemporaryDirectory() as tempdir:
            path = str(Path(tempdir) / "empty.csv")
            open(path, "a").close()

            self.assertRaises(ValueError, Table.from_csv, path)

    def test_to_csv_zip(self):
        try:
            # Test using the to_csv() method
            self.tbl.to_csv("myzip.zip")
            tmp = zip_archive.unzip_archive("myzip.zip")
            assert_matching_tables(self.tbl, Table.from_csv(tmp))

            # Test using the to_csv_zip() method
            self.tbl.to_zip_csv("myzip.zip")
            tmp = zip_archive.unzip_archive("myzip.zip")
            assert_matching_tables(self.tbl, Table.from_csv(tmp))
        finally:
            os.unlink("myzip.zip")

    def test_to_civis(self):
        # Not really sure the best way to do this at the moment.
        pass

    def test_to_from_json(self):
        with TemporaryDirectory() as tempdir:
            path = str(Path(tempdir) / "test.json")
            self.tbl.to_json(path)

            result_tbl = Table.from_json(path)
            assert_matching_tables(self.tbl, result_tbl)

    def test_to_from_json_compressed(self):
        with TemporaryDirectory() as tempdir:
            path = str(Path(tempdir) / "test.json.gz")
            self.tbl.to_json(path)

            result_tbl = Table.from_json(path)
            assert_matching_tables(self.tbl, result_tbl)

    def test_to_from_temp_json(self):
        path = self.tbl.to_json()
        result_tbl = Table.from_json(path)
        assert_matching_tables(self.tbl, result_tbl)

    def test_to_from_temp_json_compressed(self):
        path = self.tbl.to_json(temp_file_compression="gzip")
        result_tbl = Table.from_json(path)
        assert_matching_tables(self.tbl, result_tbl)

    def test_to_from_json_line_delimited(self):
        with TemporaryDirectory() as tempdir:
            path = str(Path(tempdir) / "test.json")
            self.tbl.to_json(path, line_delimited=True)

            result_tbl = Table.from_json(path, line_delimited=True)
            assert_matching_tables(self.tbl, result_tbl)

    def test_to_from_json_line_delimited_compressed(self):
        with TemporaryDirectory() as tempdir:
            path = str(Path(tempdir) / "test.json.gz")
            self.tbl.to_json(path, line_delimited=True)

            result_tbl = Table.from_json(path, line_delimited=True)
            assert_matching_tables(self.tbl, result_tbl)

    def test_columns(self):
        # Test that columns are listed correctly
        self.assertEqual(self.tbl.columns, ["first", "last"])

    def test_add_column(self):
        # Test that a new column is added correctly
        self.tbl.add_column("middle", index=1)
        self.assertEqual(self.tbl.columns[1], "middle")

    def test_column_add_dupe(self):
        # Test that we can't add an existing column name
        self.assertRaises(ValueError, self.tbl.add_column, "first")

    def test_add_column_if_exists(self):
        self.tbl.add_column("first", if_exists="replace")
        self.assertEqual(self.tbl.columns, ["first", "last"])

    def test_remove_column(self):
        # Test that column is removed correctly
        self.tbl.remove_column("first")
        self.assertNotEqual(self.tbl.data[0], "first")

    def test_rename_column(self):
        # Test that you can rename a column
        self.tbl.rename_column("first", "f")
        self.assertEqual(self.tbl.columns[0], "f")

    def test_column_rename_dupe(self):
        # Test that we can't rename to a column that already exists
        self.assertRaises(ValueError, self.tbl.rename_column, "last", "first")

    def test_rename_columns(self):
        # Test renaming columns with a valid column_map
        column_map = {"first": "firstname", "last": "lastname"}
        self.tbl.rename_columns(column_map)
        self.assertEqual(self.tbl.columns, ["firstname", "lastname"])

    def test_rename_columns_partial(self):
        # Test renaming only some columns
        column_map = {"first": "firstname"}
        self.tbl.rename_columns(column_map)
        self.assertEqual(self.tbl.columns, ["firstname", "last"])

    def test_rename_columns_nonexistent(self):
        # Test renaming a column that doesn't exist
        column_map = {"nonexistent": "newname"}
        self.assertRaises(KeyError, self.tbl.rename_columns, column_map)

    def test_rename_columns_empty(self):
        # Test renaming with an empty column_map
        column_map = {}
        self.tbl.rename_columns(column_map)
        self.assertEqual(self.tbl.columns, ["first", "last"])

    def test_rename_columns_duplicate(self):
        # Test renaming to a column name that already exists
        column_map = {"first": "last"}
        self.assertRaises(ValueError, self.tbl.rename_columns, column_map)

    def test_fill_column(self):
        # Test that the column is filled
        tbl = Table(self.lst)

        # Fixed Value
        tbl.fill_column("c", 0)
        self.assertEqual(list(tbl.table["c"]), [0] * tbl.num_rows)

        # Calculated Value
        tbl.fill_column("c", lambda x: x["b"] * 2)
        self.assertEqual(list(tbl.table["c"]), [x["b"] * 2 for x in self.lst])

    def test_fillna_column(self):
        # Test that None values in the column are filled

        self.lst = [
            {"a": 1, "b": 2, "c": 3},
            {"a": 4, "b": 5, "c": None},
            {"a": 7, "b": 8, "c": 9},
            {"a": 10, "b": 11, "c": None},
            {"a": 13, "b": 14, "c": 15},
        ]

        # Fixed Value only
        tbl = Table(self.lst)
        tbl.fillna_column("c", 0)
        self.assertEqual(list(tbl.table["c"]), [3, 0, 9, 0, 15])

    def test_move_column(self):
        # Test moving a column from end to front
        self.tbl.move_column("last", 0)
        self.assertEqual(self.tbl.columns[0], "last")

    def test_convert_column(self):
        # Test that column updates
        self.tbl.convert_column("first", "upper")
        self.assertEqual(self.tbl[0], {"first": "BOB", "last": "Smith"})

    def test_convert_columns_to_str(self):
        # Test that all columns are string
        mixed_raw = [
            {"col1": 1, "col2": 2, "col3": 3},
            {"col1": "one", "col2": 2, "col3": [3, "three", 3.0]},
            {"col1": {"one": 1, "two": 2.0}, "col2": None, "col3": "three"},
        ]
        tbl = Table(mixed_raw)
        tbl.convert_columns_to_str()

        cols = tbl.get_columns_type_stats()
        type_set = {i for x in cols for i in x["type"]}
        self.assertTrue("str" in type_set and len(type_set) == 1)

    def test_convert_table(self):
        # Test that the table updates
        self.tbl.convert_table("upper")
        self.assertEqual(self.tbl[0], {"first": "BOB", "last": "SMITH"})

    def test_coalesce_columns(self):
        # Test coalescing into an existing column
        test_raw = [
            {"first": "Bob", "last": "Smith", "lastname": None},
            {"first": "Jane", "last": "", "lastname": "Doe"},
            {"first": "Mary", "last": "Simpson", "lastname": "Peters"},
        ]
        tbl = Table(test_raw)
        tbl.coalesce_columns("last", ["last", "lastname"])

        expected = Table(
            [
                {"first": "Bob", "last": "Smith"},
                {"first": "Jane", "last": "Doe"},
                {"first": "Mary", "last": "Simpson"},
            ]
        )
        assert_matching_tables(tbl, expected)

        # Test coalescing into a new column
        tbl = Table(test_raw)
        tbl.coalesce_columns("new_last", ["last", "lastname"])
        expected = Table(
            [
                {"first": "Bob", "new_last": "Smith"},
                {"first": "Jane", "new_last": "Doe"},
                {"first": "Mary", "new_last": "Simpson"},
            ]
        )
        assert_matching_tables(tbl, expected)

    def test_unpack_dict(self):
        test_dict = [{"a": 1, "b": {"nest1": 1, "nest2": 2}}]
        test_table = Table(test_dict)

        # Test that dict at the top level
        test_table.unpack_dict("b", prepend=False)
        self.assertEqual(test_table.columns, ["a", "nest1", "nest2"])

    def test_unpack_list(self):
        test_table = Table([{"a": 1, "b": [1, 2, 3]}])

        # Test that list at the top level
        test_table.unpack_list("b", replace=True)
        self.assertEqual(["a", "b_0", "b_1", "b_2"], test_table.columns)

    def test_unpack_list_with_mixed_col(self):
        # Test unpacking column with non-list items
        mixed_tbl = Table([{"id": 1, "tag": [1, 2, None, 4]}, {"id": 2, "tag": None}])
        tbl_unpacked = Table(mixed_tbl.unpack_list("tag"))

        # Make sure result has the right number of columns
        self.assertEqual(len(tbl_unpacked.columns), 5)

        result_table = Table(
            [
                {"id": 1, "tag_0": 1, "tag_1": 2, "tag_2": None, "tag_3": 4},
                {"id": 2, "tag_0": None, "tag_1": None, "tag_2": None, "tag_3": None},
            ]
        )

        # Check that the values for both rows are distributed correctly
        self.assertEqual(
            result_table.data[0] + result_table.data[1],
            tbl_unpacked.data[0] + tbl_unpacked.data[1],
        )

    def test_unpack_nested_columns_as_rows(self):
        # A Table with mixed content
        test_table = Table(
            [
                {"id": 1, "nested": {"A": 1, "B": 2, "C": 3}, "extra": "hi"},
                {"id": 2, "nested": {"A": 4, "B": 5, "I": 6}, "extra": "hi"},
                {"id": 3, "nested": "string!", "extra": "hi"},
                {"id": 4, "nested": None, "extra": "hi"},
                {"id": 5, "nested": ["this!", "is!", "a!", "list!"], "extra": "hi"},
            ]
        )

        standalone = test_table.unpack_nested_columns_as_rows("nested")

        # Check that the columns are as expected
        self.assertEqual(["uid", "id", "nested", "value"], standalone.columns)

        # Check that the row count is as expected
        self.assertEqual(standalone.num_rows, 11)

        # Check that the uids are unique, indicating that each row is unique
        self.assertEqual(len({row["uid"] for row in standalone}), 11)

    def test_unpack_nested_columns_as_rows_expanded(self):
        test_table = Table(
            [
                {"id": 1, "nested": {"A": 1, "B": 2, "C": 3}, "extra": "hi"},
                {"id": 2, "nested": {"A": 4, "B": 5, "I": 6}, "extra": "hi"},
                {"id": 3, "nested": "string!", "extra": "hi"},
                {"id": 4, "nested": None, "extra": "hi"},
                {"id": 5, "nested": ["this!", "is!", "a!", "list!"], "extra": "hi"},
            ]
        )

        expanded = test_table.unpack_nested_columns_as_rows("nested", expand_original=True)

        # Check that the columns are as expected
        self.assertEqual(["uid", "id", "extra", "nested", "nested_value"], expanded.columns)

        # Check that the row count is as expected
        self.assertEqual(expanded.num_rows, 12)

        # Check that the uids are unique, indicating that each row is unique
        self.assertEqual(len({row["uid"] for row in expanded}), 12)

    def test_cut(self):
        # Test that the cut works correctly
        cut_tbl = self.tbl.cut("first")
        self.assertEqual(cut_tbl.columns, ["first"])

    def test_row_select(self):
        tbl = Table([["foo", "bar", "baz"], ["c", 4, 9.3], ["a", 2, 88.2], ["b", 1, 23.3]])
        expected = Table([{"foo": "a", "bar": 2, "baz": 88.2}])

        # Try with this method
        select_tbl = tbl.select_rows("{foo} == 'a' and {baz} > 88.1")
        self.assertEqual(select_tbl.data[0], expected.data[0])

        # And try with this method
        select_tbl2 = tbl.select_rows(lambda row: row.foo == "a" and row.baz > 88.1)
        self.assertEqual(select_tbl2.data[0], expected.data[0])

    def test_remove_null_rows(self):
        # Test that null rows are removed from a single column
        null_table = Table([{"a": 1, "b": 2}, {"a": 1, "b": None}])
        self.assertEqual(null_table.remove_null_rows("b").num_rows, 1)

        # Teest that null rows are removed from multiple columns
        null_table = Table([{"a": 1, "b": 2, "c": 3}, {"a": 1, "b": None, "c": 3}])
        self.assertEqual(null_table.remove_null_rows(["b", "c"]).num_rows, 1)

    def test_long_table(self):
        # Create a long table, that is 4 rows long
        tbl = Table([{"id": 1, "tag": [1, 2, 3, 4]}])
        self.assertEqual(tbl.long_table(["id"], "tag").num_rows, 4)

        # Assert that column has been dropped
        self.assertEqual(tbl.columns, ["id"])

        # Assert that column has been retained
        tbl_keep = Table([{"id": 1, "tag": [1, 2, 3, 4]}])
        tbl_keep.long_table(["id"], "tag", retain_original=True)
        self.assertEqual(tbl_keep.columns, ["id", "tag"])

    def test_long_table_with_na(self):
        # Create a long table that is 4 rows long
        tbl = Table([{"id": 1, "tag": [1, 2, 3, 4]}, {"id": 2, "tag": None}])
        self.assertEqual(tbl.long_table(["id"], "tag").num_rows, 4)

        # Assert that column has been dropped
        self.assertEqual(tbl.columns, ["id"])

        # Assert that column has been retained
        tbl_keep = Table([{"id": 1, "tag": [1, 2, 3, 4]}, {"id": 2, "tag": None}])
        tbl_keep.long_table(["id"], "tag", retain_original=True)
        self.assertEqual(tbl_keep.columns, ["id", "tag"])

    def test_rows(self):
        # Test that there is only one row in the table
        self.assertEqual(self.tbl.num_rows, 1)

    def test_first(self):
        # Test that the first value in the table is returned.
        self.assertEqual(self.tbl.first, "Bob")

        # Test empty value returns None
        empty_tbl = Table([[1], [], [3]])
        self.assertIsNone(empty_tbl.first)

    def test_get_item(self):
        # Test indexing on table

        # Test a valid column
        tbl = Table(self.lst)
        lst = [1, 4, 7, 10, 13]
        self.assertEqual(tbl["a"], lst)

        # Test a valid row
        row = {"a": 4, "b": 5, "c": 6}
        self.assertEqual(tbl[1], row)

    def test_column_data(self):
        # Test that that the data in the column is returned as a list

        # Test a valid column
        tbl = Table(self.lst)
        lst = [1, 4, 7, 10, 13]
        self.assertEqual(tbl.column_data("a"), lst)

        # Test an invalid column
        self.assertRaises(TypeError, tbl["c"])

    def test_row_data(self):
        # Test a valid column
        tbl = Table(self.lst)
        row = {"a": 4, "b": 5, "c": 6}
        self.assertEqual(tbl.row_data(1), row)

    def test_stack(self):
        tbl1 = self.tbl.select_rows(lambda x: x)
        tbl2 = Table([{"first": "Mary", "last": "Nichols"}])
        # Different column names shouldn't matter for stack()
        tbl3 = Table([{"f": "Lucy", "l": "Peterson"}])
        tbl1.stack(tbl2, tbl3)

        expected_tbl = Table(petl.stack(self.tbl.table, tbl2.table, tbl3.table))
        assert_matching_tables(expected_tbl, tbl1)

    def test_concat(self):
        tbl1 = self.tbl.select_rows(lambda x: x)
        tbl2 = Table([{"first": "Mary", "last": "Nichols"}])
        tbl3 = Table([{"first": "Lucy", "last": "Peterson"}])
        tbl1.concat(tbl2, tbl3)

        expected_tbl = Table(petl.cat(self.tbl.table, tbl2.table, tbl3.table))
        assert_matching_tables(expected_tbl, tbl1)

    def test_chunk(self):
        test_table = Table(petl.randomtable(3, 499, seed=42))
        chunks = test_table.chunk(100)

        # Assert rows of each is 100
        for c in chunks[:3]:
            self.assertEqual(100, c.num_rows)

        # Assert last table is 99
        self.assertEqual(99, chunks[4].num_rows)

    def test_match_columns(self):
        raw = [
            {"first name": "Mary", "LASTNAME": "Nichols", "Middle__Name": "D"},
            {"first name": "Lucy", "LASTNAME": "Peterson", "Middle__Name": "S"},
        ]
        tbl = Table(raw)

        desired_raw = [
            {"first_name": "Mary", "middle_name": "D", "last_name": "Nichols"},
            {"first_name": "Lucy", "middle_name": "S", "last_name": "Peterson"},
        ]
        desired_tbl = Table(desired_raw)

        # Test with fuzzy matching
        tbl.match_columns(desired_tbl.columns)
        assert_matching_tables(desired_tbl, tbl)

        # Test disable fuzzy matching, and fail due due to the missing cols
        self.assertRaises(
            TypeError,
            Table(raw).match_columns,
            desired_tbl.columns,
            fuzzy_match=False,
            if_missing_columns="fail",
        )

        # Test disable fuzzy matching, and fail due to the extra cols
        self.assertRaises(
            TypeError,
            Table(raw).match_columns,
            desired_tbl.columns,
            fuzzy_match=False,
            if_extra_columns="fail",
        )

        # Test table that already has the right columns, shouldn't need fuzzy match
        tbl = Table(desired_raw)
        tbl.match_columns(
            desired_tbl.columns,
            fuzzy_match=False,
            if_missing_columns="fail",
            if_extra_columns="fail",
        )
        assert_matching_tables(desired_tbl, tbl)

        # Test table with missing col, verify the missing col gets added by default
        tbl = Table(
            [
                {"first name": "Mary", "LASTNAME": "Nichols"},
                {"first name": "Lucy", "LASTNAME": "Peterson"},
            ]
        )
        tbl.match_columns(desired_tbl.columns)
        desired_tbl = (
            Table(desired_raw).remove_column("middle_name").add_column("middle_name", index=1)
        )
        assert_matching_tables(desired_tbl, tbl)

        # Test table with extra col, verify the extra col gets removed by default
        tbl = Table(
            [
                {
                    "first name": "Mary",
                    "LASTNAME": "Nichols",
                    "Age": 32,
                    "Middle__Name": "D",
                },
                {
                    "first name": "Lucy",
                    "LASTNAME": "Peterson",
                    "Age": 26,
                    "Middle__Name": "S",
                },
            ]
        )
        desired_tbl = Table(desired_raw)
        tbl.match_columns(desired_tbl.columns)
        assert_matching_tables(desired_tbl, tbl)

        # Test table with two columns that normalize the same and aren't in desired cols, verify
        # they both get removed.
        tbl = Table(
            [
                {
                    "first name": "Mary",
                    "LASTNAME": "Nichols",
                    "Age": 32,
                    "Middle__Name": "D",
                    "AGE": None,
                },
                {
                    "first name": "Lucy",
                    "LASTNAME": "Peterson",
                    "Age": 26,
                    "Middle__Name": "S",
                    "AGE": None,
                },
            ]
        )
        tbl.match_columns(desired_tbl.columns)
        assert_matching_tables(desired_tbl, tbl)

        # Test table with two columns that match desired cols, verify only the first gets kept.
        tbl = Table(
            [
                {
                    "first name": "Mary",
                    "LASTNAME": "Nichols",
                    "First Name": None,
                    "Middle__Name": "D",
                },
                {
                    "first name": "Lucy",
                    "LASTNAME": "Peterson",
                    "First Name": None,
                    "Middle__Name": "S",
                },
            ]
        )
        tbl.match_columns(desired_tbl.columns)
        assert_matching_tables(desired_tbl, tbl)

    def test_to_dicts(self):
        self.assertEqual(self.lst, Table(self.lst).to_dicts())
        self.assertEqual(self.lst_dicts, self.tbl.to_dicts())

    def test_reduce_rows(self):
        table = [
            ["foo", "bar"],
            ["a", 3],
            ["a", 7],
            ["b", 2],
            ["b", 1],
            ["b", 9],
            ["c", 4],
        ]
        expected = [
            {"foo": "a", "barsum": 10},
            {"foo": "b", "barsum": 12},
            {"foo": "c", "barsum": 4},
        ]

        ptable = Table(table)

        ptable.reduce_rows(
            "foo",
            lambda key, rows: [key, sum(row[1] for row in rows)],
            ["foo", "barsum"],
        )

        self.assertEqual(expected, ptable.to_dicts())

    def test_map_columns_exact(self):
        input_tbl = Table([["fn", "ln", "MID"], ["J", "B", "H"]])

        column_map = {
            "first_name": ["fn", "first"],
            "last_name": ["last", "ln"],
            "middle_name": ["mi"],
        }

        exact_tbl = Table([["first_name", "last_name", "MID"], ["J", "B", "H"]])
        input_tbl.map_columns(column_map)
        assert_matching_tables(input_tbl, exact_tbl)

    def test_map_columns_fuzzy(self):
        input_tbl = Table([["fn", "ln", "Mi_"], ["J", "B", "H"]])

        column_map = {
            "first_name": ["fn", "first"],
            "last_name": ["last", "ln"],
            "middle_name": ["mi"],
        }

        fuzzy_tbl = Table([["first_name", "last_name", "middle_name"], ["J", "B", "H"]])
        input_tbl.map_columns(column_map, exact_match=False)
        assert_matching_tables(input_tbl, fuzzy_tbl)

    def test_get_column_max_with(self):
        tbl = Table(
            [
                ["a", "b", "c"],
                ["wide_text", False, "slightly longer text"],
                ["text", 2, "byte_text🏽‍⚕️✊🏽🤩"],
            ]
        )

        # Basic test
        self.assertEqual(tbl.get_column_max_width("a"), 9)

        # Doesn't break for non-strings
        self.assertEqual(tbl.get_column_max_width("b"), 5)

        # Evaluates based on byte length rather than char length
        self.assertEqual(tbl.get_column_max_width("c"), 33)

    def test_sort(self):
        # Test basic sort
        unsorted_tbl = Table([["a", "b"], [3, 1], [2, 2], [1, 3]])
        sorted_tbl = unsorted_tbl.sort()
        self.assertEqual(sorted_tbl[0], {"a": 1, "b": 3})

        # Test column sort
        unsorted_tbl = Table([["a", "b"], [3, 1], [2, 2], [1, 3]])
        sorted_tbl = unsorted_tbl.sort("b")
        self.assertEqual(sorted_tbl[0], {"a": 3, "b": 1})

        # Test reverse sort
        unsorted_tbl = Table([["a", "b"], [3, 1], [2, 2], [1, 3]])
        sorted_tbl = unsorted_tbl.sort(reverse=True)
        self.assertEqual(sorted_tbl[0], {"a": 3, "b": 1})

    def test_set_header(self):
        # Rename columns
        tbl = Table([["one", "two"], [1, 2], [3, 4]])
        new_tbl = tbl.set_header(["oneone", "twotwo"])

        self.assertEqual(new_tbl[0], {"oneone": 1, "twotwo": 2})

        # Change number of columns
        tbl = Table([["one", "two"], [1, 2], [3, 4]])
        new_tbl = tbl.set_header(["one"])

        self.assertEqual(new_tbl[0], {"one": 1})

    def test_bool(self):
        empty = Table()
        not_empty = Table([{"one": 1, "two": 2}])

        self.assertEqual(not empty, True)
        self.assertEqual(not not_empty, False)

    def test_use_petl(self):
        # confirm that this method doesn't exist for parsons.Table
        self.assertRaises(AttributeError, getattr, Table, "skipcomments")

        tbl = Table(
            [
                ["col1", "col2"],
                [
                    "# this is a comment row",
                ],
                ["a", 1],
                ["#this is another comment", "this is also ignored"],
                ["b", 2],
            ]
        )
        tbl_expected = Table([["col1", "col2"], ["a", 1], ["b", 2]])

        tbl_after = tbl.use_petl("skipcomments", "#")
        assert_matching_tables(tbl_expected, tbl_after)

        tbl.use_petl("skipcomments", "#", update_table=True)
        assert_matching_tables(tbl_expected, tbl)

        from petl.util.base import Table as PetlTable

        tbl_petl = tbl.use_petl("skipcomments", "#", to_petl=True)
        self.assertIsInstance(tbl_petl, PetlTable)

    def test_deduplicate(self):
        # Confirm deduplicate works with no keys for one-column duplicates
        tbl = Table([["a"], [1], [2], [2], [3]])
        tbl_expected = Table([["a"], [1], [2], [3]])
        tbl.deduplicate()
        assert_matching_tables(tbl_expected, tbl)

        # Confirm deduplicate works with no keys for multiple columns
        tbl = Table([["a", "b"], [1, 2], [1, 2], [1, 3], [2, 3]])
        tbl_expected = Table(
            [
                ["a", "b"],
                [1, 2],
                [1, 3],
                [2, 3],
            ]
        )
        tbl.deduplicate()
        assert_matching_tables(tbl_expected, tbl)

        # Confirm deduplicate works with one key for multiple columns
        tbl = Table([["a", "b"], [1, 3], [1, 2], [1, 2], [2, 3]])
        tbl_expected = Table(
            [
                ["a", "b"],
                [1, 3],
                [2, 3],
            ]
        )
        tbl.deduplicate(keys=["a"])
        assert_matching_tables(tbl_expected, tbl)

        # Confirm sorting deduplicate works with one key for multiple columns

        # Note that petl sorts on the column(s) you're deduping on
        # Meaning it will ignore the 'b' column below
        # That is, the first row, [1,3],
        # would not get moved to after the [1,2]
        tbl = Table([["a", "b"], [2, 3], [1, 3], [1, 2], [1, 2]])
        tbl_expected = Table(
            [
                ["a", "b"],
                [1, 3],
                [2, 3],
            ]
        )
        tbl.deduplicate(keys=["a"], presorted=False)
        assert_matching_tables(tbl_expected, tbl)

        # Confirm sorting deduplicate works for two of two columns
        tbl = Table([["a", "b"], [2, 3], [1, 3], [1, 2], [1, 2]])
        tbl_expected = Table(
            [
                ["a", "b"],
                [1, 2],
                [1, 3],
                [2, 3],
            ]
        )
        tbl.deduplicate(keys=["a", "b"], presorted=False)
        assert_matching_tables(tbl_expected, tbl)

        # Confirm deduplicate works for multiple keys
        tbl = Table([["a", "b", "c"], [1, 2, 3], [1, 2, 3], [1, 2, 4], [1, 3, 2], [2, 3, 4]])
        tbl_expected = Table([["a", "b", "c"], [1, 2, 3], [1, 3, 2], [2, 3, 4]])
        tbl.deduplicate(["a", "b"])
        assert_matching_tables(tbl_expected, tbl)

    def test_head(self):
        tbl = Table([["a", "b"], [1, 2], [3, 4], [5, 6], [7, 8], [9, 10]])
        tbl_expected = Table([["a", "b"], [1, 2], [3, 4]])
        tbl.head(2)
        assert_matching_tables(tbl_expected, tbl)

    def test_tail(self):
        tbl = Table([["a", "b"], [1, 2], [3, 4], [5, 6], [7, 8], [9, 10]])
        tbl_expected = Table([["a", "b"], [7, 8], [9, 10]])
        tbl.tail(2)
        assert_matching_tables(tbl_expected, tbl)



import logging
import re

from parsons.etl import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)


class Mailchimp:
    """
    Instantiate Mailchimp Class

    `Args:`
        api_key:
            The Mailchimp-provided application key. Not required if
            ``MAILCHIMP_API_KEY`` env variable set.
    `Returns:`
        Mailchimp Class
    """

    def __init__(self, api_key=None):
        self.api_key = check_env.check("MAILCHIMP_API_KEY", api_key)
        self.domain = re.findall("(?<=-).+$", self.api_key)[0]
        self.uri = f"https://{self.domain}.api.mailchimp.com/3.0/"
        self.client = APIConnector(self.uri, auth=("x", self.api_key))

    def get_lists(
        self,
        fields=None,
        exclude_fields=None,
        count=None,
        offset=None,
        before_date_created=None,
        since_date_created=None,
        before_campaign_last_sent=None,
        since_campaign_last_sent=None,
        email=None,
        sort_field=None,
        sort_dir=None,
    ):
        """
        Get a table of lists under the account based on query parameters. Note
        that argument descriptions here are sourced from Mailchimp's official
        API documentation.

        `Args:`
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            before_date_created: string
                Restrict response to lists created before the set date. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            since_date_created: string
                Restrict results to lists created after the set date. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            before_campaign_last_sent: string
                Restrict results to lists created before the last campaign send
                date. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            since_campaign_last_sent: string
                Restrict results to lists created after the last campaign send
                date. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            email: string
                Restrict results to lists that include a specific subscriber's
                email address.
            sort_field: string, can only be 'date_created' or None
                Returns files sorted by the specified field.
            sort_dir: string, can only be 'ASC', 'DESC', or None
                Determines the order direction for sorted results.

        `Returns:`
            Table Class
        """
        params = {
            "fields": fields,
            "exclude_fields": exclude_fields,
            "count": count,
            "offset": offset,
            "before_date_created": before_date_created,
            "since_date_created": since_date_created,
            "before_campaign_last_sent": before_campaign_last_sent,
            "since_campaign_last_sent": since_campaign_last_sent,
            "email": email,
            "sort_field": sort_field,
            "sort_dir": sort_dir,
        }

        response = self.client.get_request("lists", params=params)
        tbl = Table(response["lists"])
        logger.info(f"Found {tbl.num_rows} lists.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_campaigns(
        self,
        fields=None,
        exclude_fields=None,
        count=None,
        offset=None,
        type=None,
        status=None,
        before_send_time=None,
        since_send_time=None,
        before_create_time=None,
        since_create_time=None,
        list_id=None,
        folder_id=None,
        member_id=None,
        sort_field=None,
        sort_dir=None,
    ):
        """
        Get a table of campaigns under the account based on query parameters.
        Note that argument descriptions here are sourced from Mailchimp's
        official API documentation.

        `Args:`
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            type: string, can only be 'regular', 'plaintext', 'absplit', 'rss',
            'variate', or None
                The campaign type.
            status: string, can only be 'save', 'paused', 'schedule',
            'sending', 'sent', or None
                The status of the campaign.
            before_send_time: string
                Restrict the response to campaigns sent before the set time. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            since_send_time: string
                Restrict the response to campaigns sent after the set time. We
                recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            before_create_time: string
                Restrict the response to campaigns created before the set time.
                We recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            since_create_time: string
                Restrict the response to campaigns created after the set time.
                We recommend ISO 8601 time format: 2015-10-21T15:41:36+00:00.
            list_id: string
                The unique id for the list.
            folder_id: string
                The unique folder id.
            member_id: string
                Retrieve campaigns sent to a particular list member. Member ID
                is The MD5 hash of the lowercase version of the list member’s
                email address.
            sort_field: string, can only be 'create_time', 'send_time', or None
                Returns files sorted by the specified field.
            sort_dir: string, can only be 'ASC', 'DESC', or None
                Determines the order direction for sorted results.

        `Returns:`
            Table Class
        """
        params = {
            "fields": fields,
            "exclude_fields": exclude_fields,
            "count": count,
            "offset": offset,
            "type": type,
            "status": status,
            "before_send_time": before_send_time,
            "since_send_time": since_send_time,
            "before_create_time": before_create_time,
            "since_create_time": since_create_time,
            "list_id": list_id,
            "folder_id": folder_id,
            "member_id": member_id,
            "sort_field": sort_field,
            "sort_dir": sort_dir,
        }

        response = self.client.get_request("campaigns", params=params)
        tbl = Table(response["campaigns"])
        logger.info(f"Found {tbl.num_rows} campaigns.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_members(
        self,
        list_id,
        fields=None,
        exclude_fields=None,
        count=None,
        offset=None,
        email_type=None,
        status=None,
        since_timestamp_opt=None,
        before_timestamp_opt=None,
        since_last_changed=None,
        before_last_changed=None,
        unique_email_id=None,
        vip_only=False,
        interest_category_id=None,
        interest_ids=None,
        interest_match=None,
        sort_field=None,
        sort_dir=None,
        since_last_campaign=None,
        unsubscribed_since=None,
    ):
        """
        Get a table of members in a list based on query parameters. Note that
        argument descriptions here are sourced from Mailchimp's official API
        documentation.

        `Args:`
            list_id: string
                The unique ID of the list to fetch members from.
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of fields as strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            email_type: string
                The email type.
            status: string, can only be 'subscribed', 'unsubscribed',
            'cleaned', 'pending', 'transactional', 'archived', or None
                The subscriber's status.
            since_timestamp_opt: string
                Restrict results to subscribers who opted-in after the set
                timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            before_timestamp_opt: string
                Restrict results to subscribers who opted-in before the set
                timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            since_last_changed: string
                Restrict results to subscribers whose information changed after
                the set timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            before_last_changed: string
                Restrict results to subscribers whose information changed
                before the set timeframe. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.
            unique_email_id: string
                A unique identifier for the email address across all Mailchimp
                lists. This parameter can be found in any links with Ecommerce
                Tracking enabled.
            vip_only: boolean
                A filter to return only the list's VIP members. Passing true
                will restrict results to VIP list members, passing false will
                return all list members.
            interest_category_id: string
                The unique id for the interest category.
            interest_ids: list of strings
                Used to filter list members by interests. Must be accompanied
                by interest_category_id and interest_match. The value must be a
                comma separated list of interest ids present for any supplied
                interest categories.
            interest_match: string, can only be 'any', 'all', 'none', or None
                Used to filter list members by interests. Must be accompanied
                by interest_category_id and interest_ids. "any" will match a
                member with any of the interest supplied, "all" will only match
                members with every interest supplied, and "none" will match
                members without any of the interest supplied.
            sort_field: string, can only be 'timestamp_opt',
            'timestamp_signup', 'last_changed', or None
                Returns files sorted by the specified field.
            sort_dir: string, can only be 'ASC', 'DESC', or None
                Determines the order direction for sorted results.
            since_last_campaign: string
                Filter subscribers by those
                subscribed/unsubscribed/pending/cleaned since last email
                campaign send. Member status is required to use this filter.
            unsubscribed_since: string
                Filter subscribers by those unsubscribed since a specific date.
                Using any status other than unsubscribed with this filter will
                result in an error.

        `Returns:`
            Table Class
        """
        params = {
            "fields": fields,
            "exclude_fields": exclude_fields,
            "count": count,
            "offset": offset,
            "email_type": email_type,
            "status": status,
            "since_timestamp_opt": since_timestamp_opt,
            "before_timestamp_opt": before_timestamp_opt,
            "since_last_changed": since_last_changed,
            "before_last_changed": before_last_changed,
            "unqiue_email_id": unique_email_id,
            "vip_only": vip_only,
            "interest_category_id": interest_category_id,
            "interest_ids": interest_ids,
            "interest_match": interest_match,
            "sort_field": sort_field,
            "sort_dir": sort_dir,
            "since_last_campaign": since_last_campaign,
            "unsubscribed_since": unsubscribed_since,
        }

        response = self.client.get_request(f"lists/{list_id}/members", params=params)
        tbl = Table(response["members"])
        logger.info(f"Found {tbl.num_rows} members.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_campaign_emails(
        self,
        campaign_id,
        fields=None,
        exclude_fields=None,
        count=None,
        offset=None,
        since=None,
    ):
        """
        Get a table of individual emails from a campaign based on query
        parameters. Note that argument descriptions here are sourced from
        Mailchimp's official API documentation.

        `Args:`
            campaign_id: string
                The unique ID of the campaign to fetch emails from.
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.
            since: string
                Restrict results to email activity events that occur after a
                specific time. We recommend ISO 8601 time format:
                2015-10-21T15:41:36+00:00.

        `Returns:`
            Table Class
        """
        params = {
            "fields": fields,
            "exclude_fields": exclude_fields,
            "count": count,
            "offset": offset,
            "since": since,
        }

        response = self.client.get_request(f"reports/{campaign_id}/email-activity", params=params)
        tbl = Table(response["emails"])
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()

    def get_unsubscribes(
        self, campaign_id, fields=None, exclude_fields=None, count=None, offset=None
    ):
        """
        Get a table of unsubscribes associated with a campaign based on query
        parameters. Note that argument descriptions here are sourced from
        Mailchimp's official API documentation.

        `Args:`
            campaign_id: string
                The unique ID of the campaign to fetch unsubscribes from.
            fields: list of strings
                A comma-separated list of fields to return. Reference
                parameters of sub-objects with dot notation.
            exclude_fields: list of strings
                A comma-separated list of fields to exclude. Reference
                parameters of sub-objects with dot notation.
            count: int
                The number of records to return. Default value is 10. Maximum
                value is 1000.
            offset: int
                The number of records from a collection to skip. Iterating over
                large collections with this parameter can be slow. Default
                value is 0.

        `Returns:`
            Table Class
        """
        params = {
            "fields": fields,
            "exclude_fields": exclude_fields,
            "count": count,
            "offset": offset,
        }

        response = self.client.get_request(f"reports/{campaign_id}/unsubscribed", params=params)
        tbl = Table(response["unsubscribes"])
        logger.info(f"Found {tbl.num_rows} unsubscribes for {campaign_id}.")
        if tbl.num_rows > 0:
            return tbl
        else:
            return Table()



import os
import unittest
import unittest.mock as mock

import requests_mock

from parsons import VAN, Table
from parsons.utilities import cloud_storage
from test.utils import validate_list


class TestSavedLists(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    def tearDown(self):
        pass

    @requests_mock.Mocker()
    def test_get_saved_lists(self, m):
        json = {
            "count": 1,
            "items": [
                {
                    "savedListId": 517612,
                    "listCount": 974656,
                    "name": "LikelyParents(16andunder)_DWID_S... - MN",
                    "doorCount": 520709,
                    "description": "null",
                }
            ],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "savedLists", json=json)

        expected = ["savedListId", "listCount", "name", "doorCount", "description"]

        self.assertTrue(validate_list(expected, self.van.get_saved_lists()))

    @requests_mock.Mocker()
    def test_get_saved_list(self, m):
        saved_list_id = 517612

        json = {
            "savedListId": 517612,
            "listCount": 974656,
            "name": "LikelyParents(16andunder)_DWID_S... - MN",
            "doorCount": 520709,
            "description": "null",
        }

        m.get(self.van.connection.uri + f"savedLists/{saved_list_id}", json=json)

        # expected = ['savedListId', 'listCount', 'name', 'doorCount', 'description']

        self.assertEqual(self.van.get_saved_list(saved_list_id), json)

    def test_upload_saved_list(self):
        cloud_storage.post_file = mock.MagicMock()
        cloud_storage.post_file.return_value = "https://box.com/my_file.zip"

        self.van.connection._soap_client = mock.MagicMock()
        self.van.get_folders = mock.MagicMock()
        self.van.get_folders.return_value = [{"folderId": 1}]

        tbl = Table([["VANID"], ["1"], ["2"], ["3"]])
        self.van.upload_saved_list(
            tbl, "GOTV List", 1, replace=True, url_type="S3", bucket="tmc-scratch"
        )
        assert self.van.connection._soap_client.service.CreateAndStoreSavedList.called

        @requests_mock.Mocker()
        def test_upload_saved_list_rest(self):
            cloud_storage.post_file = mock.MagicMock()
            cloud_storage.post_file.return_value = "https://box.com/my_file.zip"
            self.van.get_folders = mock.MagicMock()
            self.van.get_folders.return_value = [{"folderId": 1}]

            tbl = Table([["VANID"], ["1"], ["2"], ["3"]])
            response = self.van.upload_saved_list_rest(
                tbl=tbl,
                url_type="S3",
                folder_id=1,
                list_name="GOTV List",
                description="parsons test list",
                callback_url="https://webhook.site/69ab58c3-a3a7-4ed8-828c-1ea850cb4160",
                columns=["VANID"],
                id_column="VANID",
                bucket="tmc-scratch",
                overwrite=517612,
            )
            self.assertIn("jobId", response)

    @requests_mock.Mocker()
    def test_get_folders(self, m):
        json = {
            "count": 2,
            "items": [
                {"folderId": 5046, "name": "#2018_MN_active_universe"},
                {"folderId": 2168, "name": "API Generated Lists"},
            ],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "folders", json=json)

        expected = ["folderId", "name"]

        self.assertTrue(validate_list(expected, self.van.get_folders()))

    @requests_mock.Mocker()
    def test_get_folder(self, m):
        folder_id = 5046

        json = {"folderId": 5046, "name": "#2018_MN_active_universe"}

        m.get(self.van.connection.uri + f"folders/{folder_id}", json=json)

        self.assertEqual(json, self.van.get_folder(folder_id))

    @requests_mock.Mocker()
    def test_export_job_types(self, m):
        json = {
            "count": 1,
            "items": [{"exportJobTypeId": 4, "name": "SavedListExport"}],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "exportJobTypes", json=json)

        expected = ["exportJobTypeId", "name"]

        self.assertTrue(validate_list(expected, self.van.get_export_job_types()))

    @requests_mock.Mocker()
    def test_export_job_create(self, m):
        saved_list_id = 517612

        json = {
            "status": "Completed",
            "errorCode": "null",
            "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
            "activistCodes": "null",
            "canvassFileRequestId": 448,
            "dateExpired": "2018-09-08T16:04:00Z",
            "surveyQuestions": "null",
            "webhookUrl": "https://www.nothing.com/",
            "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",
            "savedListId": 517612,
            "districtFields": "null",
            "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
            "customFields": "null",
            "type": 4,
            "exportJobId": 448,
        }

        m.post(self.van.connection.uri + "exportJobs", json=json, status_code=201)

        # expected = [
        #     'status',
        #     'errorCode',
        #     'exportJobGuid',
        #     'activistCodes',
        #     'canvassFileRequestId',
        #     'dateExpired',
        #     'surveyQuestions',
        #     'webhookUrl',
        #     'downloadUrl',
        #     'savedListId',
        #     'districtFields',
        #     'canvassFileRequestGuid',
        #     'customFields',
        #     'type',
        #     'exportJobId']

        self.assertEqual(json, self.van.export_job_create(saved_list_id))

    @requests_mock.Mocker()
    def test_get_export_job(self, m):
        export_job_id = 448

        json = {
            "status": "Completed",
            "errorCode": "null",
            "exportJobGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
            "activistCodes": "null",
            "canvassFileRequestId": 448,
            "dateExpired": "2018-09-08T16:04:00Z",
            "surveyQuestions": "null",
            "webhookUrl": "https://www.nothing.com/",
            "downloadUrl": "https://ngpvan.blob.core.windows.net/canvass-files-savedlistexport/bf4d1297-1c77-3fb2-03bd-f0acda122d37_2018-09-08T13:03:27.7191831-04:00.csv",
            "savedListId": 517612,
            "districtFields": "null",
            "canvassFileRequestGuid": "bf4d1297-1c77-3fb2-03bd-f0acda122d37",
            "customFields": "null",
            "type": 4,
            "exportJobId": 448,
        }

        # expected = [
        #     'status',
        #     'errorCode',
        #     'exportJobGuid',
        #     'activistCodes',
        #     'canvassFileRequestId',
        #     'dateExpired',
        #     'surveyQuestions',
        #     'webhookUrl',
        #     'downloadUrl',
        #     'savedListId',
        #     'districtFields',
        #     'canvassFileRequestGuid',
        #     'customFields',
        #     'type',
        #     'exportJobId']

        m.get(self.van.connection.uri + f"exportJobs/{export_job_id}", json=json)

        self.assertEqual(json, self.van.get_export_job(export_job_id))



"""Class for Box API.

To get authentication info (this eventually belongs in the docs for
this connector):

Information on Box site here:
https://developer.box.com/guides/applications/custom-apps/oauth2-setup/

1. Set up Box account.
2. Go to Developer console (https://app.box.com/developers/console).
3. Either select an existing app or create a new one. To create a new one:
  - Select "New App" > "Custom App" > "OAuth 2.0 with JWT".
4. View your app and select "Configuration" in the left menu.
5. Scroll down to get the client id & secret, and just above it
   select OAuth2.0 with JWT (Server Authentication) and
   generate a developer token, aka "access token".

"""

import logging
import tempfile

import boxsdk

from parsons.etl.table import Table
from parsons.utilities.check_env import check as check_env
from parsons.utilities.files import create_temp_file, create_temp_file_for_path

logger = logging.getLogger(__name__)

DEFAULT_FOLDER_ID = "0"


class Box(object):
    """Box is a file storage provider.

    `Args:`
        client_id: str
            Box client (account) id -- probably a 16-char alphanumeric.
            Not required if ``BOX_CLIENT_ID`` env variable is set.
        client_secret: str
            Box private key -- probably a 32-char alphanumeric.
            Not required if ``BOX_CLIENT_SECRET`` env variable is set.
        access_token: str
            Box developer access token -- probably a 32-char alphanumeric.
            Note that this is only valid for developer use only, and should not
            be used when creating and maintaining access for typical users.
            Not required if ''BOX_ACCESS_TOKEN'' env variable is set.
    `Returns:`
            Box class

    *NOTE*: All path-based methods in this class use an intermediate
    method that looks up the relevant folder/file id by successive API
    calls. This can be slow for long paths or intermediate folders
    that contain many items. If performance is an issue, please use the
    corresponding folder_id/file_id methods for each function.
    """

    # In what formats can we upload/save Tables to Box? For now csv and JSON.
    ALLOWED_FILE_FORMATS = ["csv", "json"]

    def __init__(self, client_id=None, client_secret=None, access_token=None):
        client_id = check_env("BOX_CLIENT_ID", client_id)
        client_secret = check_env("BOX_CLIENT_SECRET", client_secret)
        access_token = check_env("BOX_ACCESS_TOKEN", access_token)

        oauth = boxsdk.OAuth2(
            client_id=client_id, client_secret=client_secret, access_token=access_token
        )
        self.client = boxsdk.Client(oauth)

    def create_folder(self, path) -> str:
        """Create a Box folder.

        `Args`:
            path: str
               Path to the folder to be created. If no slashes are present,
               path will be name of folder created in the default folder.

        `Returns`:
            str: The Box id of the newly-created folder.
        """
        if "/" in path:
            parent_folder_path, folder_name = path.rsplit(sep="/", maxsplit=1)
            parent_folder_id = self.get_item_id(path=parent_folder_path)
        else:
            folder_name = path
            parent_folder_id = DEFAULT_FOLDER_ID
        return self.create_folder_by_id(folder_name, parent_folder_id=parent_folder_id)

    def create_folder_by_id(self, folder_name, parent_folder_id=DEFAULT_FOLDER_ID) -> str:
        """Create a Box folder.

        `Args`:
            folder_name: str
               The name to give to the new folder.
            parent_folder_id: str
               Folder id of the parent folder in which to create the new folder. If
               omitted, the default folder will be used.

        `Returns`:
            str: The Box id of the newly-created folder.
        """
        subfolder = self.client.folder(parent_folder_id).create_subfolder(folder_name)
        return subfolder.id

    def delete_folder(self, path) -> None:
        """Delete a Box folder.

        `Args`:
            folder_id: str
               Path to the folder to delete.
        """
        folder_id = self.get_item_id(path)
        self.delete_folder_by_id(folder_id=folder_id)

    def delete_folder_by_id(self, folder_id) -> None:
        """Delete a Box folder.

        `Args`:
            folder_id: str
               The Box id of the folder to delete.
        """
        self.client.folder(folder_id=folder_id).delete()

    def delete_file(self, path) -> None:
        """Delete a Box file.

        `Args`:
            path: str
              Path to the file to delete.
        """
        file_id = self.get_item_id(path)
        self.delete_file_by_id(file_id=file_id)

    def delete_file_by_id(self, file_id) -> None:
        """Delete a Box file.

        `Args`:
            file_id: str
              The Box id of the file to delete.
        """
        self.client.file(file_id=file_id).delete()

    def list(self, path="", item_type=None) -> Table:
        """Return a Table of Box files and/or folders found at a path.

        `Args`:
            path:str
               If specified, the slash-separated path of the folder to be listed.
               If omitted, the default folder will be used.
            item_type: str
               Optionally which type of items should be returned, typically either
               `file` or `folder`. If omitted, all items will be returned.

        `Returns`: Table
            A Parsons table of items in the folder and their attributes.
        """
        if path:
            folder_id = self.get_item_id(path)
        else:
            folder_id = DEFAULT_FOLDER_ID
        return self.list_items_by_id(folder_id=folder_id, item_type=item_type)

    def list_items_by_id(self, folder_id=DEFAULT_FOLDER_ID, item_type=None) -> Table:
        url = "https://api.box.com/2.0/folders/" + folder_id
        json_response = self.client.make_request("GET", url)

        items = Table(json_response.json()["item_collection"]["entries"])
        if item_type:
            items = items.select_rows(lambda row: row.type == item_type)
        return items

    def list_files_by_id(self, folder_id=DEFAULT_FOLDER_ID) -> Table:
        """List all Box files in a folder.

        `Args`:
            folder_id: str
               The Box id of the folder in which to search. If omitted,
               search in the default folder.
        `Returns`: Table
            A Parsons table of files and their attributes.
        """
        return self.list_items_by_id(folder_id=folder_id, item_type="file")

    def list_folders_by_id(self, folder_id=DEFAULT_FOLDER_ID) -> Table:
        """List all Box folders.

        `Args`:
            folder_id: str
               The Box id of the folder in which to search. If omitted,
               search in the default folder.
        `Returns`: Table
            A Parsons table of folders and their attributes.
        """
        return self.list_items_by_id(folder_id=folder_id, item_type="folder")

    def upload_table(self, table, path="", format="csv") -> boxsdk.object.file.File:
        """Save the passed table to Box.

        `Args`:
            table:Table
               The Parsons table to be saved.
            path: str
               Optionally, file path to filename where table should be saved.
            format: str
               For now, only 'csv' and 'json'; format in which to save table.

        `Returns`: BoxFile
            A Box File object
        """
        if "/" in path:
            folder_path, file_name = path.rsplit(sep="/", maxsplit=1)
            folder_id = self.get_item_id(path=folder_path)
        else:  # pragma: no cover
            file_name = path
            folder_id = DEFAULT_FOLDER_ID

        return self.upload_table_to_folder_id(
            table=table, file_name=file_name, folder_id=folder_id, format=format
        )

    def upload_table_to_folder_id(
        self, table, file_name, folder_id=DEFAULT_FOLDER_ID, format="csv"
    ) -> boxsdk.object.file.File:
        """Save the passed table to Box.

        `Args`:
            table:Table
               The Parsons table to be saved.
            file_name: str
               The filename under which it should be saved in Box.
            folder_id: str
               Optionally, the id of the subfolder in which it should be saved.
            format: str
               For now, only 'csv' and 'json'; format in which to save table.

        `Returns`: BoxFile
            A Box File object
        """

        if format not in self.ALLOWED_FILE_FORMATS:
            raise ValueError(
                f"Format argument to upload_table() must be in one "
                f'of {self.ALLOWED_FILE_FORMATS}; found "{format}"'
            )

        # Create a temp directory in which we will let Parsons create a
        # file. Both will go away automatically when we leave scope.
        with tempfile.TemporaryDirectory() as temp_dir_name:
            temp_file_path = temp_dir_name + "/table.tmp"
            if format == "csv":
                table.to_csv(local_path=temp_file_path)
            elif format == "json":
                table.to_json(local_path=temp_file_path)
            else:
                raise SystemError(
                    f'Got (theoretically) impossible format option "{format}"'
                )  # pragma: no cover

            new_file = self.client.folder(folder_id).upload(
                file_path=temp_file_path, file_name=file_name
            )
        return new_file

    def download_file(self, path: str, local_path: str = None) -> str:
        """Download a Box object to a local file.

        `Args`:
            path: str
                The slash-separated path to the file in Box.
            local_path: str
                The local path where the file will be downloaded. If not
                specified, a temporary file will be created and returned, and
                that file will be removed automatically when the script is done
                running.

        `Returns:`
            str
                The path of the new file
        """
        if not local_path:
            # Temp file will be around as long as enclosing process is running,
            # which we need, because the Table we return will continue to use it.
            local_path = create_temp_file_for_path(path)

        file_id = self.get_item_id(path)

        with open(local_path, "wb") as output_file:
            self.client.file(file_id).download_to(output_file)

        return local_path

    def get_table(self, path, format="csv") -> Table:
        """Get a table that has been saved to Box in csv or JSON format.

        `Args`:
            path: str
                The slash-separated path to the file containing the table.
            format: str
                 Format in which Table has been saved; for now, only 'csv' or 'json'.

        `Returns`: Table
            A Parsons Table.
        """
        file_id = self.get_item_id(path)
        return self.get_table_by_file_id(file_id=file_id, format=format)

    def get_table_by_file_id(self, file_id, format="csv") -> Table:
        """Get a table that has been saved to Box in csv or JSON format.

        `Args`:
            file_id: str
                The Box file_id of the table to be retrieved.
            format: str
                 Format in which Table has been saved; for now, only 'csv' or 'json'.

        `Returns`: Table
            A Parsons Table.
        """
        if format not in self.ALLOWED_FILE_FORMATS:
            raise ValueError(
                f"Format argument to upload_table() must be in one "
                f'of {self.ALLOWED_FILE_FORMATS}; found "{format}"'
            )

        # Temp file will be around as long as enclosing process is running,
        # which we need, because the Table we return will continue to use it.
        output_file_name = create_temp_file()
        with open(output_file_name, "wb") as output_file:
            self.client.file(file_id).download_to(output_file)

        if format == "csv":
            return Table.from_csv(output_file_name)
        elif format == "json":
            return Table.from_json(output_file_name)
        else:
            raise SystemError(
                f'Got (theoretically) impossible format option "{format}"'
            )  # pragma: no cover

    def get_item_id(self, path, base_folder_id=DEFAULT_FOLDER_ID) -> str:
        """Given a path-like object, try to return the id for the file or
        folder at the end of the path.

        *NOTE*: This method makes one API call for each level in
        `path`, so can be slow for long paths or intermediate folders
        containing very many items.

        `Args`:
            path: str
                A slash-separated path from the base folder to the file or
                folder in question.
            base_folder_id: str
                 What to use as the base folder for the path. By default, use
                 the default folder.

        `Returns`: Table
            A Parsons Table.

        """
        try:
            # Grab the leftmost element in the path - this is what we're
            # looking for in this folder.
            if "/" in path:
                this_element, path = path.split(sep="/", maxsplit=1)
                if path == "":
                    raise ValueError('Illegal trailing "/" in file path')

            else:
                this_element = path
                path = ""

            # Look in our current base_folder for an item whose name matches the
            # current element. If we're at initial, non-recursed call, base_folder
            # will be default folder.
            item_id = None
            for item in self.client.folder(folder_id=base_folder_id).get_items():
                if item.name == this_element:
                    item_id = item.id
                    break

            if item_id is None:
                raise ValueError(f'No file or folder named "{this_element}"')

            # If there are no more elements left in path, this is the item we're after.
            if not len(path):
                return item_id

            # If there *are* more elements in the path, we need to check that this item is
            # in fact a folder so we can recurse and search inside it.
            if item.type != "folder":
                raise ValueError(f'Invalid folder "{this_element}"')

            return self.get_item_id(path=path, base_folder_id=item_id)

        # At the top level of the recursion, when we have the entire path, we want
        # to attach it to the error message. If caught at some inner level of the
        # recursion, just pass it on up.
        except ValueError as e:
            if base_folder_id == DEFAULT_FOLDER_ID:
                raise ValueError(f'{e}: "{path}"') from e
            else:
                raise



import logging
import os
import pickle
from contextlib import contextmanager

import mysql.connector as mysql
import petl

from parsons import Table
from parsons.databases.alchemy import Alchemy
from parsons.databases.database_connector import DatabaseConnector
from parsons.databases.mysql.create_table import MySQLCreateTable
from parsons.databases.table import BaseTable
from parsons.utilities import check_env, files

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000

logger = logging.getLogger(__name__)


class MySQL(DatabaseConnector, MySQLCreateTable, Alchemy):
    """
    Connect to a MySQL database.

    `Args:`
        username: str
            Required if env variable ``MYSQL_USERNAME`` not populated
        password: str
            Required if env variable ``MYSQL_PASSWORD`` not populated
        host: str
            Required if env variable ``MYSQL_HOST`` not populated
        db: str
            Required if env variable ``MYSQL_DB`` not populated
        port: int
            Can be set by env variable ``MYSQL_PORT`` or argument.
    """

    def __init__(self, host=None, username=None, password=None, db=None, port=3306):
        super().__init__()

        self.username = check_env.check("MYSQL_USERNAME", username)
        self.password = check_env.check("MYSQL_PASSWORD", password)
        self.host = check_env.check("MYSQL_HOST", host)
        self.db = check_env.check("MYSQL_DB", db)
        self.port = port or os.environ.get("MYSQL_PORT")

    @contextmanager
    def connection(self):
        """
        Generate a MySQL connection. The connection is set up as a python "context manager", so
        it will be closed automatically (and all queries committed) when the connection goes out
        of scope.

        When using the connection, make sure to put it in a ``with`` block (necessary for
        any context manager):
        ``with mysql.connection() as conn:``

        `Returns:`
            MySQL `connection` object
        """

        # Create a mysql connection and cursor
        connection = mysql.connect(
            host=self.host,
            user=self.username,
            passwd=self.password,
            database=self.db,
            port=self.port,
        )

        try:
            yield connection
        except mysql.Error:
            connection.rollback()
            raise
        else:
            connection.commit()
        finally:
            connection.close()

    @contextmanager
    def cursor(self, connection):
        cur = connection.cursor(buffered=True)

        try:
            yield cur
        finally:
            cur.close()

    def query(self, sql, parameters=None):
        """
        Execute a query against the database. Will return ``None`` if the query returns zero rows.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the `mysql style <https://security.openstack.org/guidelines/dg_parameterize-database-queries.html>`_.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

            # Note that the name contains a quote, which could break your query if not escaped
            # properly.
            name = "Beatrice O'Brady"
            sql = "SELECT * FROM my_table WHERE name = %s"
            mysql.query(sql, parameters=[name])

        .. code-block:: python

            names = ["Allen Smith", "Beatrice O'Brady", "Cathy Thompson"]
            placeholders = ', '.join('%s' for item in names)
            sql = f"SELECT * FROM my_table WHERE name IN ({placeholders})"
            mysql.query(sql, parameters=names)

        `Args:`
            sql: str
                A valid SQL statement
            parameters: list
                A list of python variables to be converted into SQL values in your query

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.

        """

        with self.connection() as connection:
            return self.query_with_connection(sql, connection, parameters=parameters)

    def query_with_connection(self, sql, connection, parameters=None, commit=True):
        """
        Execute a query against the database, with an existing connection. Useful for batching
        queries together. Will return ``None`` if the query returns zero rows.

        `Args:`
            sql: str
                A valid SQL statement
            connection: obj
                A connection object obtained from ``mysql.connection()``
            parameters: list
                A list of python variables to be converted into SQL values in your query
            commit: boolean
                Whether to commit the transaction immediately. If ``False`` the transaction will
                be committed when the connection goes out of scope and is closed (or you can
                commit manually with ``connection.commit()``).

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        with self.cursor(connection) as cursor:
            # The python connector can only execute a single sql statement, so we will
            # break up each statement and execute them separately.
            for s in sql.strip().split(";"):
                if len(s) != 0:
                    logger.debug(f"SQL Query: {sql}")
                    cursor.execute(s, parameters)

            if commit:
                connection.commit()

            # If the SQL query provides no response, then return None
            if not cursor.description:
                logger.debug("Query returned 0 rows")
                return None

            else:
                # Fetch the data in batches, and "pickle" the rows to a temp file.
                # (We pickle rather than writing to, say, a CSV, so that we maintain
                # all the type information for each field.)
                temp_file = files.create_temp_file()

                with open(temp_file, "wb") as f:
                    # Grab the header
                    pickle.dump(cursor.column_names, f)

                    while True:
                        batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                        if len(batch) == 0:
                            break

                        logger.debug(f"Fetched {len(batch)} rows.")
                        for row in batch:
                            pickle.dump(row, f)

                # Load a Table from the file
                final_tbl = Table(petl.frompickle(temp_file))

                logger.debug(f"Query returned {final_tbl.num_rows} rows.")
                return final_tbl

    def copy(
        self,
        tbl: Table,
        table_name: str,
        if_exists: str = "fail",
        chunk_size: int = 1000,
        strict_length: bool = True,
    ):
        """
        Copy a :ref:`parsons-table` to the database.

        .. note::
            This method utilizes extended inserts rather `LOAD DATA INFILE` since
            many MySQL Database configurations do not allow data files to be
            loaded. It results in a minor performance hit compared to `LOAD DATA`.

        `Args:`
            tbl: parsons.Table
                A Parsons table object
            table_name: str
                The destination schema and table (e.g. ``my_schema.my_table``)
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            chunk_size: int
                The number of rows to insert per query.
            strict_length: bool
                If the database table needs to be created, strict_length determines whether
                the created table's column sizes will be sized to exactly fit the current data,
                or if their size will be rounded up to account for future values being larger
                then the current dataset. defaults to ``True``
        """

        if tbl.num_rows == 0:
            logger.info("Parsons table is empty. Table will not be created.")
            return None

        with self.connection() as connection:
            # Create table if not exists
            if self._create_table_precheck(connection, table_name, if_exists):
                sql = self.create_statement(tbl, table_name, strict_length=strict_length)
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f"Table {table_name} created.")

            # Chunk tables in batches of 1K rows, though this can be tuned and
            # optimized further.
            for t in tbl.chunk(chunk_size):
                sql = self._insert_statement(t, table_name)
                self.query_with_connection(sql, connection, commit=False)

    def _insert_statement(self, tbl, table_name):
        """
        Convert the table data into a string for bulk importing.
        """

        # Single column tables
        if len(tbl.columns) == 1:
            values = [f"({row[0]})" for row in tbl.data]

        # Multi-column tables
        else:
            values = [str(row) for row in tbl.data]

        # Create full insert statement
        sql = f"""INSERT INTO {table_name}
                  ({",".join(tbl.columns)})
                  VALUES {",".join(values)};"""

        return sql

    def _create_table_precheck(self, connection, table_name, if_exists):
        """
        Helper to determine what to do when you need a table that may already exist.

        `Args:`
            connection: obj
                A connection object obtained from ``mysql.connection()``
            table_name: str
                The table to check
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``,
                or ``truncate`` the table.
        `Returns:`
            bool
                True if the table needs to be created, False otherwise.
        """

        if if_exists not in ["fail", "truncate", "append", "drop"]:
            raise ValueError("Invalid value for `if_exists` argument")

        # If the table exists, evaluate the if_exists argument for next steps.
        if self.table_exists(table_name):
            if if_exists == "fail":
                raise ValueError("Table already exists.")

            if if_exists == "truncate":
                sql = f"TRUNCATE TABLE {table_name}"
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f"{table_name} truncated.")
                return False

            if if_exists == "drop":
                sql = f"DROP TABLE {table_name}"
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f"{table_name} dropped.")
                return True

        else:
            return True

    def table_exists(self, table_name: str) -> bool:
        """
        Check if a table or view exists in the database.

        `Args:`
            table_name: str
                The table name

        `Returns:`
            boolean
                ``True`` if the table exists and ``False`` if it does not.
        """

        if self.query(f"SHOW TABLES LIKE '{table_name}'").first == table_name:
            return True
        else:
            return False

    def table(self, table_name):
        # Return a BaseTable table object

        return BaseTable(self, table_name)



import os
import unittest

import requests_mock

from parsons import VAN
from test.utils import validate_list

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestNGPVAN(unittest.TestCase):
    def setUp(self):
        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters")

    def tearDown(self):
        pass

    @requests_mock.Mocker()
    def test_get_events(self, m):
        json = {
            "count": 6,
            "items": [
                {
                    "eventId": 1062,
                    "startDate": "2010-05-25T11:00:00-05:00",
                    "codes": "null",
                    "endDate": "2010-05-25T15:00:00-05:00",
                    "name": "Sample",
                    "roles": "null",
                    "isOnlyEditableByCreatingUser": "true",
                    "ticketCategories": "null",
                    "eventType": {"eventTypeId": 29166, "name": "Confirmation Calls"},
                    "notes": "null",
                    "districtFieldValue": "null",
                    "locations": "null",
                    "shifts": "null",
                    "voterRegistrationBatches": "null",
                    "createdDate": "2010-05-25T11:55:00Z",
                    "financialProgram": "null",
                    "shortName": "Sample",
                    "isPubliclyViewable": "null",
                    "isActive": "true",
                    "description": "This is a sample",
                }
            ],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "events", json=json)

        # Expected Structure
        expected = [
            "eventId",
            "startDate",
            "codes",
            "endDate",
            "name",
            "roles",
            "isOnlyEditableByCreatingUser",
            "ticketCategories",
            "eventType",
            "notes",
            "districtFieldValue",
            "locations",
            "shifts",
            "voterRegistrationBatches",
            "createdDate",
            "financialProgram",
            "shortName",
            "isPubliclyViewable",
            "isActive",
            "description",
        ]

        self.assertTrue(validate_list(expected, self.van.get_events()))

    @requests_mock.Mocker()
    def test_get_event(self, m):
        event_id = 1062

        json = {
            "eventId": 1062,
            "startDate": "2010-05-25T11:00:00-05:00",
            "codes": "null",
            "endDate": "2010-05-25T15:00:00-05:00",
            "name": "Sample",
            "roles": "null",
            "isOnlyEditableByCreatingUser": "true",
            "ticketCategories": "null",
            "eventType": {"eventTypeId": 29166, "name": "Confirmation Calls"},
            "notes": "null",
            "districtFieldValue": "null",
            "locations": "null",
            "shifts": "null",
            "voterRegistrationBatches": "null",
            "createdDate": "2010-05-25T11:55:00Z",
            "financialProgram": "null",
            "shortName": "Sample",
            "isPubliclyViewable": "null",
            "isActive": "true",
            "description": "This is a sample",
        }

        m.get(self.van.connection.uri + "events/{}".format(event_id), json=json)

        self.assertEqual(json, self.van.get_event(event_id))

    @requests_mock.Mocker()
    def test_create_event(self, m):
        m.post(self.van.connection.uri + "events", json=750000984, status_code=204)

        # Test that it doesn't throw and error
        r = self.van.create_event(
            "Canvass 01",
            "Can01",
            "2016-06-01",
            "2016-06-02",
            296199,
            [259236],
            publicly_viewable="True",
            editable=False,
        )

        self.assertEqual(r, 750000984)

    @requests_mock.Mocker()
    def test_get_event_types(self, m):
        json = [
            {
                "eventTypeId": 296199,
                "name": "Block Party",
                "canHaveMultipleShifts": False,
                "canHaveMultipleLocations": False,
                "canHaveGoals": False,
                "canHaveRoleMaximums": False,
                "canHaveRoleMinimums": False,
                "canBeRepeatable": False,
                "roles": [
                    {"roleId": 259236, "name": "Attendee", "isEventLead": False},
                    {"roleId": 259235, "name": "Supporter", "isEventLead": False},
                    {"roleId": 259234, "name": "Volunteer", "isEventLead": False},
                ],
                "statuses": [
                    {"statusId": 4, "name": "Invited"},
                    {"statusId": 18, "name": "Left Msg"},
                    {"statusId": 14, "name": "Tentative"},
                    {"statusId": 3, "name": "Declined"},
                    {"statusId": 11, "name": "Confirmed"},
                    {"statusId": 23, "name": "Conf Twice"},
                    {"statusId": 2, "name": "Completed"},
                    {"statusId": 15, "name": "Walk In"},
                    {"statusId": 6, "name": "No Show"},
                    {"statusId": 29, "name": "Texted"},
                ],
                "color": "#7F7F7F",
                "isAtLeastOneLocationRequired": False,
                "defaultLocation": None,
                "isSharedWithMasterCommitteeByDefault": False,
                "isSharedWithChildCommitteesByDefault": False,
                "isOnlineActionsAvailable": False,
            }
        ]

        m.get(self.van.connection.uri + "events/types", json=json)

        expected = [
            "eventTypeId",
            "name",
            "canHaveMultipleShifts",
            "canHaveMultipleLocations",
            "canHaveGoals",
            "canHaveRoleMaximums",
            "canHaveRoleMinimums",
            "canBeRepeatable",
            "roles",
            "statuses",
            "color",
            "isAtLeastOneLocationRequired",
            "defaultLocation",
            "isSharedWithMasterCommitteeByDefault",
            "isSharedWithChildCommitteesByDefault",
            "isOnlineActionsAvailable",
        ]

        self.assertTrue(validate_list(expected, self.van.get_event_types()))
